from typing import Tuple

WORKFLOW_EXEC_ID_META_KEY = "workflow_exec_id"
WORKFLOW_NODE_ID_META_KEY = "workflow_node_id"

WORKFLOW_ATTRIBUTION_META_KEYS: Tuple[str, str] = (
    WORKFLOW_EXEC_ID_META_KEY,
    WORKFLOW_NODE_ID_META_KEY,
)


__all__ = []
