# 本文件由 `scripts/gen-project-constants.py` 自动生成,请勿手改.
# 数据来源: `pyproject.toml` 的 `([project] + [tool.scalim.*])`.

CLI_NAME = "scalim-cli"
DIST_NAME = "scalim"
ENV_BENCH_MAX_WORKERS = "SCALIM_BENCH_MAX_WORKERS"
ENV_BENCH_SCALE = "SCALIM_BENCH_SCALE"
ENV_BENCH_SCOPE = "SCALIM_BENCH_SCOPE"
ENV_SYNC_ANALYSIS_SCRIPT = "SCALIM_ANALYSIS_SCRIPT"
ENV_SYNC_ANALYZE_ARGS = "SCALIM_ANALYZE_ARGS"
ENV_SYNC_ANALYZE_ARGS_PREFIX = "SCALIM_ANALYZE_ARGS_"
ENV_SYNC_DEST_DIR = "SCALIM_DEST_DIR"
ENV_SYNC_LOG_DEST_DIR = "SCALIM_LOG_DEST_DIR"
ENV_SYNC_LOG_REMOTE_DIR = "SCALIM_LOG_REMOTE_DIR"
ENV_SYNC_REMOTE_HOST = "SCALIM_REMOTE_HOST"
ENV_SYNC_REPORT_ROOT_DIR = "SCALIM_REPORT_ROOT_DIR"
ENV_SYNC_TMP_REMOTE_DIR = "SCALIM_TMP_REMOTE_DIR"
IMPORT_ROOT = "scalim"
VERSION = "0.6.0"
VIZ_ARTIFACTS_ROOT = "artifacts/scalim-viz"
VIZ_DIR_NAME = "scalim-viz"
VIZ_REPLAY_ROUTE = "/__scalim_viz__/replay"

__all__ = []
